;-----------------------------------------------------------------------------
; Copyright (c) 2006-2007, mike.dld
;-----------------------------------------------------------------------------
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
; DEALINGS IN THE SOFTWARE.
;-----------------------------------------------------------------------------
;
; 2007-05-19 (mike.dld)
;   bug-fixes:
;     - last char still wasn't read correctly
;     - digits of number were reversed when using ini.set_int
;   changes:
;     - all non-public functions now start with ini.aux.*
;     - added ini.enum_sections and ini.enum_keys
;     - removed ini.query_sec (use ini.enum_* instead)
;
;-----------------------------------------------------------------------------

format MS COFF

public EXPORTS

include 'proc32.inc'
include 'macros.inc'
include 'libcommon.inc'
purge section;mov,add,sub

section '.flat' code readable align 16

mem.alloc   dd ?
mem.free    dd ?
mem.realloc dd ?
dll.load    dd ?

ini.MAX_NAME_LEN  = 1024
ini.MAX_VALUE_LEN = 4096
ini.MEM_SIZE	  = 4096
ini.BLOCK_SIZE	  = ini.MEM_SIZE / 2

struct INIFILE
  fh  dd ?
  buf dd ?
  cnt dd ?
ends

;-----------------------------------------------------------------------------
proc lib_init ;///////////////////////////////////////////////////////////////
;-----------------------------------------------------------------------------
	mov	[mem.alloc],eax
	mov	[mem.free],ebx
	mov	[mem.realloc],ecx
	mov	[dll.load],edx

	invoke	dll.load,@IMPORT
	or	eax,eax
	jz	.ok

	or	eax,-1
	ret

  .ok:	xor	eax,eax
	ret
endp

;-----------------------------------------------------------------------------
proc ini.aux.get_char f ;/////////////////////////////////////////////////////
;-----------------------------------------------------------------------------
	mov	ecx,[f]
	dec	[ecx+INIFILE.cnt]
	jns	@f
	stdcall ini.aux.preload_block,[f]
	dec	[ecx+INIFILE.cnt]
    @@: lodsb
	ret
endp

;-----------------------------------------------------------------------------
proc ini.aux.skip_spaces f ;//////////////////////////////////////////////////
;-----------------------------------------------------------------------------
	mov	ecx,[f]
    @@: stdcall ini.aux.get_char,[f]
	cmp	al,32
	je	@b
	cmp	al,13
	je	@b
	cmp	al,10
	je	@b
	cmp	al,9
	je	@b
    @@: dec	esi
	inc	[ecx+INIFILE.cnt]
	ret
endp

;-----------------------------------------------------------------------------
proc ini.aux.skip_line f ;////////////////////////////////////////////////////
;-----------------------------------------------------------------------------
	mov	ecx,[f]
    @@: stdcall ini.aux.get_char,[f]
	or	al,al
	jz	@f
	cmp	al,13
	je	@f
	cmp	al,10
	jne	@b
    @@: dec	esi
	inc	[ecx+INIFILE.cnt]
	ret
endp

;-----------------------------------------------------------------------------
proc ini.aux.preload_block f ;////////////////////////////////////////////////
;-----------------------------------------------------------------------------
	push	eax ebx ecx
	mov	ebx,[f]
	mov	esi,[ebx+INIFILE.buf]
	push	edi
	mov	edi,esi
	mov	ecx,ini.BLOCK_SIZE/4
	xor	eax,eax
	rep	stosd
	pop	edi
	invoke	file.read,[ebx+INIFILE.fh],esi,ini.BLOCK_SIZE
	mov	esi,[ebx+INIFILE.buf]
	cmp	eax,ini.BLOCK_SIZE
	jl	@f
	dec	eax
    @@: mov	[ebx+INIFILE.cnt],eax;ini.BLOCK_SIZE-1
	pop	ecx ebx eax
	ret
endp

; f_info - contains current file block number
; esi    - position in block from where to shift
; ecx    - number of bytes to shift by

;-----------------------------------------------------------------------------
proc ini.aux.shift_content f,delta ;//////////////////////////////////////////
;-----------------------------------------------------------------------------
; Shift file content starting from cursor position (~ delete)
; Content is copied by 'delta' bytes up/down
;
; If operation succeeded,
;   eax = 0
; Otherwise,
;   eax = -1
;-----------------------------------------------------------------------------
locals
  buf dd ?
endl

	cmp	[delta],0
	je	.skip

	push	ebx ecx
	invoke	mem.alloc,ini.BLOCK_SIZE
	or	eax,eax
	jz	.fail
	mov	[buf],eax

	cmp	[delta],0
	jl	.down

	mov	ebx,[f]
	mov	ecx,[ebx+INIFILE.cnt]
	mov	ebx,[ebx+INIFILE.fh]
	invoke	file.tell,ebx
	push	eax
	sub	eax,ecx
	dec	eax
	invoke	file.seek,ebx,SEEK_SET,eax
    @@: invoke	file.seek,ebx,SEEK_CUR,[delta]
	invoke	file.eof?,ebx
	or	eax,eax
	jnz	.done
	invoke	file.read,ebx,[buf],ini.BLOCK_SIZE
	mov	ecx,eax
	mov	eax,[delta]
	neg	eax
	sub	eax,ecx;ini.BLOCK_SIZE
	invoke	file.seek,ebx,SEEK_CUR,eax
	invoke	file.write,ebx,[buf],ecx;ini.BLOCK_SIZE
	jmp	@b
  .done:
	mov	eax,[delta]
	neg	eax
	invoke	file.seek,ebx,SEEK_CUR,eax
	invoke	file.seteof,ebx
;       pop     eax
	invoke	file.seek,ebx,SEEK_SET;,eax
	invoke	mem.free,[buf]
	pop	ecx ebx
  .skip:
	ret
  .fail:
	or	eax,-1
	pop	ecx ebx
	ret

  .down:
	neg	[delta]

	mov	ebx,[f]
	mov	ecx,[ebx+INIFILE.cnt]
	mov	ebx,[ebx+INIFILE.fh]
	invoke	file.tell,ebx
	push	eax
	sub	eax,ecx
	lea	edx,[eax-1]
	push	edx
    @@: invoke	file.seek,ebx,SEEK_SET,edx
	invoke	file.eof?,ebx
	or	eax,eax
	jnz	@f
	add	edx,ini.BLOCK_SIZE
	jmp	@b
    @@: cmp	edx,[esp]
	je	.skip.2
	add	edx,-ini.BLOCK_SIZE
	cmp	edx,[esp]
	jl	@f
	invoke	file.seek,ebx,SEEK_SET,edx
	invoke	file.read,ebx,[buf],ini.BLOCK_SIZE
	mov	ecx,eax
	mov	eax,[delta]
	sub	eax,ecx
	invoke	file.seek,ebx,SEEK_CUR,eax
	invoke	file.write,ebx,[buf],ecx
	jmp	@b
    @@:
  .skip.2:
	add	esp,4
;       mov     eax,[delta]
;       neg     eax
;       invoke  file.seek,ebx,SEEK_CUR,eax
;       pop     eax
	invoke	file.seek,ebx,SEEK_SET;,eax
	invoke	mem.free,[buf]
	pop	ecx ebx
	ret
endp

;-----------------------------------------------------------------------------
proc ini.aux.get_value_length f ;/////////////////////////////////////////////
;-----------------------------------------------------------------------------
	push	ebx ecx edx eax
	mov	ebx,[f]
	invoke	file.tell,[ebx+INIFILE.fh]
	push	esi [ebx+INIFILE.cnt] eax
	sub	eax,[ebx+INIFILE.cnt]
	mov	edx,eax

	stdcall ini.aux.skip_line,[f]
	invoke	file.tell,[ebx+INIFILE.fh]
	sub	eax,[ebx+INIFILE.cnt]
	sub	eax,edx
	mov	[esp+4*3],eax

	pop	eax
	add	eax,-ini.BLOCK_SIZE
	invoke	file.seek,[ebx+INIFILE.fh],SEEK_SET,eax
	stdcall ini.aux.preload_block,[f]
	pop	[ebx+INIFILE.cnt] esi
	pop	eax edx ecx ebx
	ret
endp

;-----------------------------------------------------------------------------
proc ini.aux.string_copy ;////////////////////////////////////////////////////
;-----------------------------------------------------------------------------
    @@: lodsb
	or	al,al
	jz	@f
	stosb
	jmp	@b
    @@: ret
endp

;-----------------------------------------------------------------------------
proc ini.aux.find_next_section f ;////////////////////////////////////////////
;-----------------------------------------------------------------------------
	push	ebx edi

    @@: stdcall ini.aux.skip_spaces,[f]
	cmp	al,'['
	je	@f
	or	al,al
	jz	.exit_error
	stdcall ini.aux.skip_line,[f]
	or	al,al
	jz	.exit_error
	jmp	@b
    @@:
	pop	edi ebx
	xor	eax,eax
	ret

  .exit_error:
	pop	edi ebx
	or	eax,-1
	ret
endp

;-----------------------------------------------------------------------------
proc ini.aux.find_section f,sec_name ;////////////////////////////////////////
;-----------------------------------------------------------------------------
; Find section in file
; Search is performed from the beginning of file
;
; If section is found,
;   eax = 0
;   f.pos = cursor position (right after ']' char)
; Otherwise,
;   eax = -1
;   f.pos = cursor position (at the end of file)
;-----------------------------------------------------------------------------
	push	ebx edi

	mov	ecx,[f]
	invoke	file.seek,[ecx+INIFILE.fh],SEEK_SET,0
	stdcall ini.aux.preload_block,[f]

  .next_section:
	stdcall ini.aux.find_next_section,[f]
	or	eax,eax
	jnz	.exit_error

	inc	esi
	dec	[ecx+INIFILE.cnt]
	mov	edi,[sec_name]
    @@: stdcall ini.aux.get_char,[f]
	cmp	al,']'
	je	@f
	or	al,al
	jz	.exit_error
	cmp	al,13
	je	.next_section
	cmp	al,10
	je	.next_section
	scasb
	je	@b
	jmp	.next_section
    @@:
	cmp	byte[edi],0
	jne	.next_section
	pop	edi ebx
	xor	eax,eax
	ret

  .exit_error:
	pop	edi ebx
	or	eax,-1
	ret
endp

;-----------------------------------------------------------------------------
proc ini.aux.find_key f,key_name ;////////////////////////////////////////////
;-----------------------------------------------------------------------------
; Find key in section
; Search is performed within current section starting from cursor position
;
; If key is found,
;   eax = 0
;   f.pos = cursor position (right after '=' char)
; Otherwise,
;   eax = -1
;   f.pos = cursor position (either at the end of file or right before '[' char)
;-----------------------------------------------------------------------------
	push	ebx edi

  .next_value:
	mov	edi,[key_name]
	stdcall ini.aux.skip_line,[f]
	stdcall ini.aux.skip_spaces,[f]
	or	al,al
	jz	.exit_error
	cmp	al,'['
	je	.exit_error
    @@: stdcall ini.aux.get_char,[f]
	or	al,al
	jz	.exit_error
	cmp	al,'='
	je	@f
	scasb
	je	@b
	jmp	.next_value
    @@:
	cmp	byte[edi],0
	jne	.next_value

	pop	edi ebx
	xor	eax,eax
	ret

  .exit_error:
	pop	edi ebx
	or	eax,-1
	ret
endp

;-----------------------------------------------------------------------------
proc ini.enum_sections f_name,callback ;//////////////////////////////////////
;-----------------------------------------------------------------------------
; callback(f_name, sec_name)
;-----------------------------------------------------------------------------
locals
  f	  INIFILE
  f_addr  dd ?
  sec_buf dd ?
endl

	push	ebx esi edi

	cld

	invoke	mem.alloc,ini.MAX_NAME_LEN
	or	eax,eax
	jz	.exit_error.2
	mov	[sec_buf],eax

	xor	eax,eax
	mov	[f.fh],eax
	mov	[f.buf],eax
	invoke	file.open,[f_name],O_READ
	cmp	eax,32
	jb	.exit_error
	mov	[f.fh],eax
	invoke	mem.alloc,ini.MEM_SIZE
	or	eax,eax
	jz	.exit_error
	mov	[f.buf],eax
	lea	ebx,[f]
	mov	[f_addr],ebx

	invoke	file.seek,[f.fh],SEEK_SET,0
	stdcall ini.aux.preload_block,[f_addr]

  .next_section:
	stdcall ini.aux.find_next_section,[f_addr]
	or	eax,eax
	jnz	.exit_error

	inc	esi
	dec	[f.cnt]
	mov	edi,[sec_buf]
    @@: stdcall ini.aux.get_char,[f_addr]
	cmp	al,']'
	je	@f
	or	al,al
	jz	.exit_ok
	cmp	al,13
	je	.next_section
	cmp	al,10
	je	.next_section
	stosb
	jmp	@b
    @@:
	xor	al,al
	stosb
	pushad
	mov	eax,[f_addr]
	stdcall [callback],[f_name],[sec_buf]
	or	eax,eax
	popad
	jnz	.next_section

  .exit_ok:
	invoke	file.close,[f.fh]
	invoke	mem.free,[f.buf]
	invoke	mem.free,[sec_buf]
	xor	eax,eax
	pop	edi esi ebx
	ret

  .exit_error:
	invoke	file.close,[f.fh]
	invoke	mem.free,[f.buf]
	invoke	mem.free,[sec_buf]
  .exit_error.2:
	or	eax,-1
	pop	edi esi ebx
	ret
endp

;-----------------------------------------------------------------------------
proc ini.enum_keys f_name,sec_name,callback ;/////////////////////////////////
;-----------------------------------------------------------------------------
; callback(f_name, sec_name, key_name)
;-----------------------------------------------------------------------------
locals
  f	  INIFILE
  f_addr  dd ?
  key_buf dd ?
endl

	push	ebx esi edi

	cld

	invoke	mem.alloc,ini.MAX_NAME_LEN
	or	eax,eax
	jz	.exit_error.2
	mov	[key_buf],eax

	xor	eax,eax
	mov	[f.fh],eax
	mov	[f.buf],eax
	invoke	file.open,[f_name],O_READ
	cmp	eax,32
	jb	.exit_error
	mov	[f.fh],eax
	invoke	mem.alloc,ini.MEM_SIZE
	or	eax,eax
	jz	.exit_error
	mov	[f.buf],eax
	lea	ebx,[f]
	mov	[f_addr],ebx
	stdcall ini.aux.find_section,ebx,[sec_name]
	or	eax,eax
	jnz	.exit_error

  .next_key:
	stdcall ini.aux.skip_line,[f]
	stdcall ini.aux.skip_spaces,[f]
	or	al,al
	jz	.exit_error
	cmp	al,'['
	je	.exit_error
	mov	edi,[key_buf]
    @@: stdcall ini.aux.get_char,[f]
	or	al,al
	jz	.exit_error
	cmp	al,'='
	je	@f
	stosb
	jmp	@b
    @@:
	xor	al,al
	stosb
	add	edi,-2
    @@: cmp	byte[edi],32
	ja	@f
	mov	byte[edi],0
	dec	edi
	jmp	@b
    @@: pushad
	mov	eax,[f_addr]
	stdcall [callback],[f_name],[sec_name],[key_buf]
	or	eax,eax
	popad
	jnz	.next_key

    @@: invoke	file.close,[f.fh]
	invoke	mem.free,[f.buf]
	xor	eax,eax
	stosb
	pop	edi esi ebx
	ret

  .exit_error:
	invoke	file.close,[f.fh]
	invoke	mem.free,[f.buf]
	invoke	mem.free,[key_buf]
  .exit_error.2:
	or	eax,-1
	pop	edi esi ebx
	ret
endp

;-----------------------------------------------------------------------------
proc ini.get_str f_name,sec_name,key_name,buffer,buf_len,def_val ;////////////
;-----------------------------------------------------------------------------
locals
  f	 INIFILE
  f_addr dd ?
endl

	push	ebx esi edi

	xor	eax,eax
	mov	[f.fh],eax
	mov	[f.buf],eax
	invoke	file.open,[f_name],O_READ
	cmp	eax,32
	jb	.exit_error
	mov	[f.fh],eax
	invoke	mem.alloc,ini.MEM_SIZE
	or	eax,eax
	jz	.exit_error
	mov	[f.buf],eax
	lea	ebx,[f]
	mov	[f_addr],ebx
	stdcall ini.aux.find_section,ebx,[sec_name]
	or	eax,eax
	jnz	.exit_error

	stdcall ini.aux.find_key,ebx,[key_name]
	or	eax,eax
	jnz	.exit_error

	mov	edi,[buffer]
    @@: dec	[buf_len]
	jz	@f
	stdcall ini.aux.get_char,[f_addr]
	or	al,al
	jz	@f
	cmp	al,13
	je	@f
	cmp	al,10
	je	@f
	stosb
	jmp	@b
    @@: invoke	file.close,[f.fh]
	invoke	mem.free,[f.buf]
	xor	eax,eax
	stosb
	pop	edi esi ebx
	ret

  .exit_error:
	invoke	file.close,[f.fh]
	invoke	mem.free,[f.buf]
	mov	esi,[def_val]
	mov	edi,[buffer]
    @@: lodsb
	stosb
	or	al,al
	jnz	@b
	or	eax,-1
	pop	edi esi ebx
	ret
endp

;-----------------------------------------------------------------------------
proc ini.set_str f_name,sec_name,key_name,buffer,buf_len ;////////////////////
;-----------------------------------------------------------------------------
locals
  f	 INIFILE
  f_addr dd ?
endl

	push	ebx esi edi

	xor	eax,eax
	mov	[f.fh],eax
	mov	[f.buf],eax
	invoke	file.open,[f_name],O_READ+O_WRITE+O_CREATE
	cmp	eax,32
	jb	.exit_error
	mov	[f.fh],eax
	invoke	mem.alloc,ini.MEM_SIZE
	or	eax,eax
	jz	.exit_error
	mov	[f.buf],eax
	lea	ebx,[f]
	mov	[f_addr],ebx

	stdcall ini.aux.find_section,ebx,[sec_name]
	or	eax,eax
	jnz	.create_section

	stdcall ini.aux.find_key,ebx,[key_name]
	or	eax,eax
	jnz	.create_key

  .modify_key:
	stdcall ini.aux.get_value_length,[f_addr]
	sub	eax,[buf_len]
	stdcall ini.aux.shift_content,[f_addr],eax

  .modify_key.ex:
	invoke	file.tell,[f.fh]
	sub	eax,[f.cnt]
	dec	eax
	invoke	file.seek,[f.fh],SEEK_SET,eax
	invoke	file.write,[f.fh],[buffer],[buf_len]

	pop	edi esi ebx
	xor	eax,eax
	ret

  .create_key:
	mov	edi,[f.buf]
	add	edi,ini.BLOCK_SIZE
	push	edi

  .create_key.ex:
	;mov     word[edi],0x0A0D
	;add     edi,2
	mov	esi,[key_name]
	call	ini.aux.string_copy
	mov	byte[edi],'='
	inc	edi
	mov	esi,[buffer]
	mov	ecx,[buf_len]
	rep	movsb
	mov	word[edi],0x0A0D
	add	edi,2
	mov	eax,edi

	pop	edi
	sub	eax,edi
	mov	[buffer],edi
	mov	[buf_len],eax
	neg	eax
	stdcall ini.aux.shift_content,[f_addr],eax

	jmp	.modify_key.ex

  .create_section:
	mov	edi,[f.buf]
	add	edi,ini.BLOCK_SIZE
	push	edi

	mov	esi,[sec_name]
	;mov     dword[edi],0x0A0D+('[' shl 16)
	;add     edi,3
	mov	byte[edi],'['
	inc	edi
	call	ini.aux.string_copy
	;mov     byte[edi],']'
	;inc     edi
	mov	dword[edi],']' + (0x0A0D shl 8)
	add	edi,3

	jmp	.create_key.ex

  .exit_error:
	pop	edi esi ebx
	or	eax,-1
	ret
endp

;-----------------------------------------------------------------------------
proc ini.get_int f_name,sec_name,key_name,def_val ;///////////////////////////
;-----------------------------------------------------------------------------
locals
  f	 INIFILE
  f_addr dd ?
endl

	push	ebx esi edi

	xor	eax,eax
	mov	[f.fh],eax
	mov	[f.buf],eax
	invoke	file.open,[f_name],O_READ
	cmp	eax,32
	jb	.exit_error
	mov	[f.fh],eax
	invoke	mem.alloc,ini.MEM_SIZE
	or	eax,eax
	jz	.exit_error
	mov	[f.buf],eax
	lea	ebx,[f]
	mov	[f_addr],ebx
	stdcall ini.aux.find_section,ebx,[sec_name]
	or	eax,eax
	jnz	.exit_error

	stdcall ini.aux.find_key,ebx,[key_name]
	or	eax,eax
	jnz	.exit_error

	stdcall ini.aux.skip_spaces,[f]
	xor	eax,eax
	xor	ebx,ebx
	xor	edx,edx
	stdcall ini.aux.get_char,[f_addr]
	cmp	al,'-'
	jne	.lp1
	inc	bh
    @@: stdcall ini.aux.get_char,[f_addr]
  .lp1: cmp	al,'0'
	jb	@f
	cmp	al,'9'
	ja	@f
	inc	bl
	add	eax,-'0'
	imul	edx,10
	add	edx,eax
	jmp	@b
    @@:
	or	bl,bl
	jz	.exit_error
	or	bh,bh
	jz	@f
	neg	edx
    @@: invoke	file.close,[f.fh]
	invoke	mem.free,[f.buf]
	mov	eax,edx
	pop	edi esi ebx
	ret

  .exit_error:
	invoke	file.close,[f.fh]
	invoke	mem.free,[f.buf]
	mov	eax,[def_val]
	pop	edi esi ebx
	ret
endp

;-----------------------------------------------------------------------------
proc ini.set_int f_name,sec_name,key_name,val ;///////////////////////////////
;-----------------------------------------------------------------------------
locals
  buf rb 16
endl

	push	ecx edx edi

	lea	edi,[buf]
	add	edi,15
	mov	eax,[val]
	or	eax,eax
	jns	@f
	mov	byte[edi],'-'
	neg	eax
	inc	edi
    @@: mov	ecx,10
    @@: xor	edx,edx
	idiv	ecx
	add	dl,'0'
	mov	[edi],dl
	dec	edi
	or	eax,eax
	jnz	@b
	lea	eax,[buf]
	add	eax,15
	sub	eax,edi
	inc	edi

	stdcall ini.set_str,[f_name],[sec_name],[key_name],edi,eax

	pop	edi edx ecx
	ret
endp

align 16
@IMPORT:

library \
	libfile,'libio.obj'

import	libfile, \
	file.size  ,'file.size'  ,\ ; f_name
	file.open  ,'file.open'  ,\ ; f_name  f_mode
	file.read  ,'file.read'  ,\ ; f_descr buffer   buf_len
	file.write ,'file.write' ,\ ; f_descr buffer   buf_len
	file.seek  ,'file.seek'  ,\ ; f_descr f_origin f_where
	file.eof?  ,'file.eof?'  ,\ ; f_descr
	file.seteof,'file.seteof',\ ; f_descr
	file.tell  ,'file.tell'  ,\ ; f_descr
	file.close ,'file.close'    ; f_descr

align 16
EXPORTS:

export	\
	lib_init	 ,'lib_init'	     ,\
	0x00030003	 ,'version'	     ,\
	ini.enum_sections,'ini.enum_sections',\
	ini.enum_keys	 ,'ini.enum_keys'    ,\
	ini.get_str	 ,'ini.get_str'      ,\
	ini.get_int	 ,'ini.get_int'      ,\
	ini.set_str	 ,'ini.set_str'      ,\
	ini.set_int	 ,'ini.set_int'
