;-----------------------------------------------------------------------------
; Copyright (c) 2006-2007, mike.dld
;-----------------------------------------------------------------------------
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
; DEALINGS IN THE SOFTWARE.
;-----------------------------------------------------------------------------

format MS COFF

public EXPORTS

include 'proc32.inc'
include 'macros.inc'
include 'libcommon.inc'
purge section;mov,add,sub

section '.flat' code readable align 16

mem.alloc   dd ?
mem.free    dd ?
mem.realloc dd ?
dll.load    dd ?

struct FindFileBlock
  Header    FileInfoHeader
  Info	    FileInfo
  InfoBlock FileInfoBlock
ends

struct InternalFileInfo
  Mode	   dd ?
  Position dd ?
  FileName rb 260
ends

;-----------------------------------------------------------------------------
proc lib_init ;///////////////////////////////////////////////////////////////
;-----------------------------------------------------------------------------
	mov	[mem.alloc],eax
	mov	[mem.free],ebx
	mov	[mem.realloc],ecx
	mov	[dll.load],edx
	xor	eax,eax
	ret
endp

;-----------------------------------------------------------------------------
proc file.find_first f_dir,f_mask,f_attr ;////////////////////////////////////
;-----------------------------------------------------------------------------
	ret
endp

;-----------------------------------------------------------------------------
proc file.find_next ;/////////////////////////////////////////////////////////
;-----------------------------------------------------------------------------
	ret
endp

;-----------------------------------------------------------------------------
proc file.find_close f_descr ;////////////////////////////////////////////////
;-----------------------------------------------------------------------------
	ret
endp

;-----------------------------------------------------------------------------
proc file.size f_name ;///////////////////////////////////////////////////////
;-----------------------------------------------------------------------------
locals
  f_info FileInfoBlock
endl

	lea	ebx,[f_info]
	invoke	mem.alloc,40
	push	eax
	mov	[ebx+FileInfoBlock.Function],5
	mov	[ebx+FileInfoBlock.Buffer],eax
	mov	byte[ebx+FileInfoBlock.FileName-1],0
	mov	eax,[f_name]
	mov	[ebx+FileInfoBlock.FileName],eax
	mcall	70
	pop	ebx
	push	eax
	mov	eax,ebx
	mov	ebx,dword[ebx+FileInfo.FileSize]
	invoke	mem.free,eax
	pop	eax
	ret
endp

;-----------------------------------------------------------------------------
proc file.open f_name,f_mode ;////////////////////////////////////////////////
;-----------------------------------------------------------------------------
locals
  f_info FileInfoBlock
  buf rb 40
endl

	push	ebx esi edi

	xor	ebx,ebx
	invoke	mem.alloc,sizeof.InternalFileInfo
	or	eax,eax
	jz	.exit_error
	mov	ebx,eax
	push	[f_mode]
	pop	[ebx+InternalFileInfo.Mode]
	mov	[ebx+InternalFileInfo.Position],0
	lea	edi,[ebx+InternalFileInfo.FileName]
	mov	esi,[f_name]
	mov	ecx,260/4
	cld
	rep	movsd

  .get_info:
	push	ebx
	mov	[f_info.Function],5
	lea	eax,[buf]
	mov	[f_info.Buffer],eax
	mov	byte[f_info.FileName-1],0
	mov	eax,[f_name]
	mov	[f_info.FileName],eax
	lea	ebx,[f_info]
	mcall	70
	pop	ebx
	or	eax,eax
	jz	@f
	cmp	eax,6
	jne	.exit_error.ex
    @@:
	mov	eax,ebx
	pop	edi esi ebx
	ret

  .exit_error.ex:
	test	[f_mode],O_CREATE
	jz	.exit_error
	push	ebx
	mov	[f_info.Function],2
	xor	eax,eax
	mov	[f_info.Position],eax
	mov	[f_info.Flags],eax
	mov	[f_info.Count],eax
	lea	ebx,[f_info]
	mcall	70
	pop	ebx
	or	eax,eax
	jz	.get_info

  .exit_error:
	push	eax
	invoke	mem.free,ebx
	pop	eax
	pop	edi esi ebx
	ret
endp

;-----------------------------------------------------------------------------
proc file.read f_descr,buffer,buf_len ;///////////////////////////////////////
;-----------------------------------------------------------------------------
locals
  f_info FileInfoBlock
endl

	push	ebx esi edi

	mov	ebx,[f_descr]
	test	[ebx+InternalFileInfo.Mode],O_READ
	jz	.exit_error

	xor	eax,eax
	mov	[f_info.Function],eax
	mov	[f_info.Flags],eax
	mov	byte[f_info.FileName-1],al
	push	[ebx+InternalFileInfo.Position] [buf_len] [buffer]
	pop	[f_info.Buffer] [f_info.Count] [f_info.Position]
	lea	eax,[ebx+InternalFileInfo.FileName]
	mov	[f_info.FileName],eax
	lea	ebx,[f_info]
	mcall	70
	or	eax,eax
	jz	@f
	cmp	eax,6
	jne	.exit_error
    @@:
	mov	eax,ebx
	mov	ebx,[f_descr]
	add	[ebx+InternalFileInfo.Position],eax
	pop	edi esi ebx
	ret

  .exit_error:
	or	eax,-1
	pop	edi esi ebx
	ret
endp

;-----------------------------------------------------------------------------
proc file.write f_descr,buffer,buf_len ;//////////////////////////////////////
;-----------------------------------------------------------------------------
locals
  f_info FileInfoBlock
endl

	push	ebx esi edi

	mov	ebx,[f_descr]
	test	[ebx+InternalFileInfo.Mode],O_WRITE
	jz	.exit_error

	stdcall file.eof?,[f_descr]
	or	eax,eax
	js	.exit_error
	jz	@f
	stdcall file.truncate,[f_descr]
    @@:
	mov	[f_info.Function],3
	xor	eax,eax
	mov	[f_info.Flags],eax
	mov	byte[f_info.FileName-1],al
	push	[ebx+InternalFileInfo.Position] [buf_len] [buffer]
	pop	[f_info.Buffer] [f_info.Count] [f_info.Position]
	lea	eax,[ebx+InternalFileInfo.FileName]
	mov	[f_info.FileName],eax
	lea	ebx,[f_info]
	mcall	70
	or	eax,eax
	jnz	.exit_error
    @@:
	mov	eax,ebx
	mov	ebx,[f_descr]
	add	[ebx+InternalFileInfo.Position],eax
	pop	edi esi ebx
	ret

  .exit_error:
	or	eax,-1
	pop	edi esi ebx
	ret
endp

;-----------------------------------------------------------------------------
proc file.seek f_descr,f_origin,f_where ;/////////////////////////////////////
;-----------------------------------------------------------------------------
	push	ebx ecx edx

	mov	ecx,[f_descr]
	lea	eax,[ecx+InternalFileInfo.FileName]
	stdcall file.size,eax
	or	eax,eax
	jnz	.exit_error
	mov	edx,[f_where]
	cmp	[f_origin],SEEK_SET
	jne	.n_set
	mov	[ecx+InternalFileInfo.Position],edx
	jmp	.exit_ok

  .n_set:
	cmp	[f_origin],SEEK_CUR
	jne	.n_cur
	add	[ecx+InternalFileInfo.Position],edx
	jmp	.exit_ok

  .n_cur:
	cmp	[f_origin],SEEK_END
	jne	.exit_error
	neg	edx
	add	edx,ebx
	mov	[ecx+InternalFileInfo.Position],edx

  .exit_ok:

	cmp	[ecx+InternalFileInfo.Position],0
	jge	@f
	mov	[ecx+InternalFileInfo.Position],0
    @@:
;       cmp     ebx,[ecx+InternalFileInfo.Position]
;       jae     @f
;       mov     [ecx+InternalFileInfo.Position],ebx
;   @@:
	xor	eax,eax
	pop	edx ecx ebx
	ret

  .exit_error:
	or	eax,-1
	pop	edx ecx ebx
	ret
endp

;-----------------------------------------------------------------------------
proc file.eof? f_descr ;//////////////////////////////////////////////////////
;-----------------------------------------------------------------------------
	push	ebx ecx

	mov	ecx,[f_descr]
	lea	eax,[ecx+InternalFileInfo.FileName]
	stdcall file.size,eax
	or	eax,eax
	jnz	.exit_error

	xor	eax,eax
	cmp	[ecx+InternalFileInfo.Position],ebx
	jb	@f
	inc	eax
    @@: pop	ecx ebx
	ret

  .exit_error:
	or	eax,-1
	pop	ecx ebx
	ret
endp

file.seteof equ file.truncate
;-----------------------------------------------------------------------------
proc file.truncate f_descr ;//////////////////////////////////////////////////
;-----------------------------------------------------------------------------
locals
  f_info FileInfoBlock
endl

	push	ebx esi edi

	mov	ebx,[f_descr]
	test	[ebx+InternalFileInfo.Mode],O_WRITE
	jz	.exit_error

	mov	[f_info.Function],4
	mov	eax,[ebx+InternalFileInfo.Position]
	mov	[f_info.Position],eax
	xor	eax,eax
	mov	[f_info.Flags],eax
	mov	[f_info.Count],eax
	mov	[f_info.Buffer],eax
	mov	byte[f_info.FileName-1],al
	lea	eax,[ebx+InternalFileInfo.FileName]
	mov	[f_info.FileName],eax
	lea	ebx,[f_info]
	mcall	70
	cmp	eax,2
	je	.exit_error
	cmp	eax,8
	je	.exit_error
    @@:
	pop	edi esi ebx
	ret

  .exit_error:
	or	eax,-1
	pop	edi esi ebx
	ret
endp

;-----------------------------------------------------------------------------
proc file.tell f_descr ;//////////////////////////////////////////////////////
;-----------------------------------------------------------------------------
	mov	eax,[f_descr]
	mov	eax,[eax+InternalFileInfo.Position]
	ret
endp

;-----------------------------------------------------------------------------
proc file.close f_descr ;/////////////////////////////////////////////////////
;-----------------------------------------------------------------------------
	mov	eax,[f_descr]
	mov	[eax+InternalFileInfo.Mode],0
	mov	[eax+InternalFileInfo.FileName],0
	invoke	mem.free,eax

	xor	eax,eax
	ret
endp

align 16
EXPORTS:

export	\
	lib_init       ,'lib_init',\
	0x00010001     ,'version',\
	file.find_first,'file.find_first',\
	file.find_next ,'file.find_next',\
	file.find_close,'file.find_close',\
	file.size      ,'file.size',\
	file.open      ,'file.open',\
	file.read      ,'file.read',\
	file.write     ,'file.write',\
	file.seek      ,'file.seek',\
	file.tell      ,'file.tell',\
	file.eof?      ,'file.eof?',\
	file.seteof    ,'file.seteof',\
	file.truncate  ,'file.truncate',\
	file.close     ,'file.close'
